# test_bundle_uri.py -- test bundle URI compatibility with CGit
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for bundle URI compatibility with CGit.

Bundle URIs were introduced in Git 2.38.
"""

import os
import socket
import tempfile
import threading
import time
from http.server import HTTPServer, SimpleHTTPRequestHandler

from dulwich.bundle import create_bundle_from_repo, write_bundle
from dulwich.bundle_uri import fetch_bundle_uri, parse_bundle_list
from dulwich.client import BundleClient
from dulwich.repo import MemoryRepo, Repo

from .utils import CompatTestCase, rmtree_ro, run_git_or_fail


class BundleURICompatTestCase(CompatTestCase):
    """Test bundle URI compatibility with CGit.

    Bundle URIs require Git 2.38 or later.
    """

    min_git_version = (2, 38, 0)

    def setUp(self) -> None:
        super().setUp()
        self.test_dir = tempfile.mkdtemp()
        self.addCleanup(rmtree_ro, self.test_dir)

    def test_bundle_list_format_compat(self) -> None:
        """Test that bundle list format is compatible with git's expectations."""
        # Create a bundle list in the format git expects
        bundle_list_content = b"""[bundle]
\tversion = 1
\tmode = all

[bundle "main"]
\turi = https://example.com/main.bundle
\tcreationToken = 1700000000
"""
        # Parse with dulwich
        bundle_list = parse_bundle_list(bundle_list_content)

        # Verify parsed correctly
        self.assertEqual(bundle_list.version, 1)
        self.assertEqual(bundle_list.mode, "all")
        self.assertEqual(len(bundle_list.entries), 1)
        self.assertEqual(bundle_list.entries[0].id, "main")
        self.assertEqual(bundle_list.entries[0].uri, "https://example.com/main.bundle")
        self.assertEqual(bundle_list.entries[0].creation_token, 1700000000)

    def test_bundle_list_any_mode_compat(self) -> None:
        """Test bundle list with 'any' mode (geographic distribution)."""
        bundle_list_content = b"""[bundle]
\tversion = 1
\tmode = any

[bundle "us-east"]
\turi = https://us-east.example.com/repo.bundle
\tlocation = us-east

[bundle "eu-west"]
\turi = https://eu-west.example.com/repo.bundle
\tlocation = eu-west
"""
        bundle_list = parse_bundle_list(bundle_list_content)

        self.assertEqual(bundle_list.mode, "any")
        self.assertEqual(len(bundle_list.entries), 2)

        locations = {e.location for e in bundle_list.entries}
        self.assertEqual(locations, {"us-east", "eu-west"})

    def test_bundle_list_with_filter_compat(self) -> None:
        """Test bundle list with partial clone filters."""
        bundle_list_content = b"""[bundle]
\tversion = 1
\tmode = all

[bundle "full"]
\turi = https://example.com/full.bundle
\tcreationToken = 1700000000

[bundle "blobless"]
\turi = https://example.com/blobless.bundle
\tcreationToken = 1700000000
\tfilter = blob:none
"""
        bundle_list = parse_bundle_list(bundle_list_content)

        self.assertEqual(len(bundle_list.entries), 2)

        full = next(e for e in bundle_list.entries if e.id == "full")
        self.assertIsNone(full.filter)

        blobless = next(e for e in bundle_list.entries if e.id == "blobless")
        self.assertEqual(blobless.filter, "blob:none")

    def test_bundle_list_with_heuristic_compat(self) -> None:
        """Test bundle list with creationToken heuristic."""
        bundle_list_content = b"""[bundle]
\tversion = 1
\tmode = all
\theuristic = creationToken

[bundle "2024-01-daily"]
\turi = https://example.com/2024-01-daily.bundle
\tcreationToken = 1704067200

[bundle "2024-01-weekly"]
\turi = https://example.com/2024-01-weekly.bundle
\tcreationToken = 1703462400

[bundle "base"]
\turi = https://example.com/base.bundle
\tcreationToken = 1700000000
"""
        bundle_list = parse_bundle_list(bundle_list_content)

        self.assertEqual(bundle_list.heuristic, "creationToken")
        self.assertEqual(len(bundle_list.entries), 3)

        # Verify creation tokens are parsed correctly
        tokens = {e.id: e.creation_token for e in bundle_list.entries}
        self.assertEqual(tokens["2024-01-daily"], 1704067200)
        self.assertEqual(tokens["2024-01-weekly"], 1703462400)
        self.assertEqual(tokens["base"], 1700000000)

    def test_bundle_from_git_fetch_with_bundle_client(self) -> None:
        """Test fetching a bundle created by git using BundleClient."""
        # Create a repository with git
        repo_path = os.path.join(self.test_dir, "source_repo")
        os.makedirs(repo_path)
        run_git_or_fail(["init"], cwd=repo_path)
        run_git_or_fail(["config", "user.name", "Test User"], cwd=repo_path)
        run_git_or_fail(["config", "user.email", "test@example.com"], cwd=repo_path)

        # Create some commits
        test_file = os.path.join(repo_path, "test.txt")
        with open(test_file, "w") as f:
            f.write("Hello, World!\n")
        run_git_or_fail(["add", "test.txt"], cwd=repo_path)
        run_git_or_fail(["commit", "-m", "Initial commit"], cwd=repo_path)

        with open(test_file, "a") as f:
            f.write("More content\n")
        run_git_or_fail(["add", "test.txt"], cwd=repo_path)
        run_git_or_fail(["commit", "-m", "Second commit"], cwd=repo_path)

        # Create bundle with git
        bundle_path = os.path.join(self.test_dir, "test.bundle")
        run_git_or_fail(
            ["bundle", "create", bundle_path, "HEAD", "master"], cwd=repo_path
        )

        # Use BundleClient to fetch from the bundle
        client = BundleClient()
        target_repo = MemoryRepo()
        self.addCleanup(target_repo.close)

        result = client.fetch(bundle_path, target_repo)

        # Verify refs were fetched
        self.assertIn(b"refs/heads/master", result.refs)

        # Verify objects were fetched
        master_sha = result.refs[b"refs/heads/master"]
        self.assertIn(master_sha, target_repo.object_store)

        # Verify commit content
        commit = target_repo.object_store[master_sha]
        self.assertEqual(commit.message, b"Second commit\n")

    def test_dulwich_bundle_readable_by_git(self) -> None:
        """Test that bundles created by dulwich can be read by git."""
        # Create a repository with dulwich
        repo_path = os.path.join(self.test_dir, "dulwich_repo")
        repo = Repo.init(repo_path, mkdir=True)
        self.addCleanup(repo.close)

        # Create objects using git (easier for creating valid commits)
        run_git_or_fail(["config", "user.name", "Test User"], cwd=repo_path)
        run_git_or_fail(["config", "user.email", "test@example.com"], cwd=repo_path)

        test_file = os.path.join(repo_path, "file.txt")
        with open(test_file, "w") as f:
            f.write("Test content\n")
        run_git_or_fail(["add", "file.txt"], cwd=repo_path)
        run_git_or_fail(["commit", "-m", "Test commit"], cwd=repo_path)

        # Re-open repo to pick up changes
        repo.close()
        repo = Repo(repo_path)
        self.addCleanup(repo.close)

        # Create bundle with dulwich
        bundle_path = os.path.join(self.test_dir, "dulwich.bundle")
        bundle = create_bundle_from_repo(repo)
        self.addCleanup(bundle.close)

        with open(bundle_path, "wb") as f:
            write_bundle(f, bundle)

        # Verify git can verify the bundle (run_git_or_fail will raise if it fails)
        run_git_or_fail(["bundle", "verify", bundle_path], cwd=repo_path)

        # Clone from the dulwich-created bundle using git - this is the real test
        clone_path = os.path.join(self.test_dir, "git_clone")
        run_git_or_fail(["clone", bundle_path, clone_path])

        # Verify clone succeeded by checking the file content
        cloned_file = os.path.join(clone_path, "file.txt")
        with open(cloned_file) as f:
            self.assertEqual(f.read(), "Test content\n")

    def test_bundle_list_relative_uris(self) -> None:
        """Test that relative URIs in bundle lists are resolved correctly."""
        bundle_list_content = b"""[bundle]
\tversion = 1
\tmode = all

[bundle "relative"]
\turi = bundles/main.bundle
\tcreationToken = 1700000000

[bundle "absolute-path"]
\turi = /repo/bundles/backup.bundle
\tcreationToken = 1699000000
"""
        base_uri = "https://example.com/git/myrepo/"
        bundle_list = parse_bundle_list(bundle_list_content, base_uri=base_uri)

        relative_entry = next(e for e in bundle_list.entries if e.id == "relative")
        self.assertEqual(
            relative_entry.uri, "https://example.com/git/myrepo/bundles/main.bundle"
        )

        absolute_entry = next(e for e in bundle_list.entries if e.id == "absolute-path")
        self.assertEqual(
            absolute_entry.uri, "https://example.com/repo/bundles/backup.bundle"
        )

    def test_incremental_bundle_with_prerequisites(self) -> None:
        """Test incremental bundles with prerequisites work correctly."""
        # Create a repository with multiple commits
        repo_path = os.path.join(self.test_dir, "incremental_repo")
        os.makedirs(repo_path)
        run_git_or_fail(["init"], cwd=repo_path)
        run_git_or_fail(["config", "user.name", "Test User"], cwd=repo_path)
        run_git_or_fail(["config", "user.email", "test@example.com"], cwd=repo_path)

        # Create base commit
        test_file = os.path.join(repo_path, "test.txt")
        with open(test_file, "w") as f:
            f.write("Line 1\n")
        run_git_or_fail(["add", "test.txt"], cwd=repo_path)
        run_git_or_fail(["commit", "-m", "Commit 1"], cwd=repo_path)

        base_commit = run_git_or_fail(["rev-parse", "HEAD"], cwd=repo_path).strip()

        # Create full bundle (base) - must include a ref
        base_bundle_path = os.path.join(self.test_dir, "base.bundle")
        run_git_or_fail(
            ["bundle", "create", base_bundle_path, "HEAD", "master"], cwd=repo_path
        )

        # Create more commits for incremental bundle
        with open(test_file, "a") as f:
            f.write("Line 2\n")
        run_git_or_fail(["add", "test.txt"], cwd=repo_path)
        run_git_or_fail(["commit", "-m", "Commit 2"], cwd=repo_path)

        with open(test_file, "a") as f:
            f.write("Line 3\n")
        run_git_or_fail(["add", "test.txt"], cwd=repo_path)
        run_git_or_fail(["commit", "-m", "Commit 3"], cwd=repo_path)

        # Create incremental bundle with prerequisite
        incremental_bundle_path = os.path.join(self.test_dir, "incremental.bundle")
        run_git_or_fail(
            [
                "bundle",
                "create",
                incremental_bundle_path,
                f"{base_commit.decode()}..HEAD",
            ],
            cwd=repo_path,
        )

        # Clone using base bundle first, then apply incremental
        clone_path = os.path.join(self.test_dir, "incremental_clone")
        run_git_or_fail(["clone", base_bundle_path, clone_path])

        # Fetch from incremental bundle
        run_git_or_fail(["fetch", incremental_bundle_path, "HEAD"], cwd=clone_path)

        # Verify both bundles were applied
        output = run_git_or_fail(["log", "--oneline", "FETCH_HEAD"], cwd=clone_path)
        # Parse log output - format is "<short-sha> <message>" per line
        log_lines = output.strip().split(b"\n")
        commit_messages = [line.split(b" ", 1)[1] for line in log_lines if b" " in line]
        # Should have all 3 commits in reverse chronological order
        self.assertEqual(commit_messages[0], b"Commit 3")
        self.assertEqual(commit_messages[1], b"Commit 2")
        self.assertEqual(commit_messages[2], b"Commit 1")


class BundleURIHTTPCompatTestCase(CompatTestCase):
    """Test bundle URI over HTTP compatibility."""

    min_git_version = (2, 38, 0)

    def setUp(self) -> None:
        super().setUp()
        self.test_dir = tempfile.mkdtemp()
        self.addCleanup(rmtree_ro, self.test_dir)
        self.server = None
        self.server_thread = None

    def tearDown(self) -> None:
        if self.server:
            self.server.shutdown()
        if self.server_thread:
            self.server_thread.join(timeout=5)
        super().tearDown()

    def _start_http_server(self, serve_dir: str) -> str:
        """Start a simple HTTP server and return its URL."""

        class QuietHandler(SimpleHTTPRequestHandler):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, directory=serve_dir, **kwargs)

            def log_message(self, format, *args):
                pass  # Suppress logging

        self.server = HTTPServer(("127.0.0.1", 0), QuietHandler)
        port = self.server.server_address[1]

        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.daemon = True
        self.server_thread.start()

        # Wait for server to be ready
        url = f"http://127.0.0.1:{port}"
        for _ in range(50):  # Try for up to 5 seconds
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.settimeout(0.1)
                sock.connect(("127.0.0.1", port))
                sock.close()
                break
            except OSError:
                time.sleep(0.1)
        else:
            raise RuntimeError("HTTP server failed to start")

        return url

    def test_fetch_bundle_from_http(self) -> None:
        """Test fetching a bundle from HTTP URL."""
        # Create a repository and bundle
        repo_path = os.path.join(self.test_dir, "source")
        os.makedirs(repo_path)
        run_git_or_fail(["init"], cwd=repo_path)
        run_git_or_fail(["config", "user.name", "Test User"], cwd=repo_path)
        run_git_or_fail(["config", "user.email", "test@example.com"], cwd=repo_path)

        test_file = os.path.join(repo_path, "README.md")
        with open(test_file, "w") as f:
            f.write("# Test Repository\n")
        run_git_or_fail(["add", "README.md"], cwd=repo_path)
        run_git_or_fail(["commit", "-m", "Initial commit"], cwd=repo_path)

        # Create bundle
        serve_dir = os.path.join(self.test_dir, "serve")
        os.makedirs(serve_dir)
        bundle_path = os.path.join(serve_dir, "repo.bundle")
        run_git_or_fail(
            ["bundle", "create", bundle_path, "HEAD", "master"], cwd=repo_path
        )

        # Start HTTP server
        base_url = self._start_http_server(serve_dir)

        # Fetch bundle using dulwich
        bundle, bundle_list = fetch_bundle_uri(f"{base_url}/repo.bundle")

        # Should return a bundle, not a bundle list
        self.assertIsNotNone(bundle)
        self.assertIsNone(bundle_list)
        self.addCleanup(bundle.close)

        # Verify bundle contents
        self.assertIn(b"refs/heads/master", bundle.references)

    def test_fetch_bundle_list_from_http(self) -> None:
        """Test fetching a bundle list from HTTP URL."""
        # Create serve directory with bundle list
        serve_dir = os.path.join(self.test_dir, "serve")
        os.makedirs(serve_dir)

        # Create a simple repository and bundle
        repo_path = os.path.join(self.test_dir, "source")
        os.makedirs(repo_path)
        run_git_or_fail(["init"], cwd=repo_path)
        run_git_or_fail(["config", "user.name", "Test User"], cwd=repo_path)
        run_git_or_fail(["config", "user.email", "test@example.com"], cwd=repo_path)

        test_file = os.path.join(repo_path, "test.txt")
        with open(test_file, "w") as f:
            f.write("test\n")
        run_git_or_fail(["add", "test.txt"], cwd=repo_path)
        run_git_or_fail(["commit", "-m", "Test"], cwd=repo_path)

        bundle_path = os.path.join(serve_dir, "main.bundle")
        run_git_or_fail(
            ["bundle", "create", bundle_path, "HEAD", "master"], cwd=repo_path
        )

        # Start HTTP server first to get URL
        base_url = self._start_http_server(serve_dir)

        # Create bundle list file
        bundle_list_content = f"""[bundle]
\tversion = 1
\tmode = all

[bundle "main"]
\turi = {base_url}/main.bundle
\tcreationToken = 1700000000
"""
        bundle_list_path = os.path.join(serve_dir, "bundle-list")
        with open(bundle_list_path, "w") as f:
            f.write(bundle_list_content)

        # Fetch bundle list
        bundle, bundle_list = fetch_bundle_uri(f"{base_url}/bundle-list")

        # Should return a bundle list, not a bundle
        self.assertIsNone(bundle)
        self.assertIsNotNone(bundle_list)

        # Verify bundle list contents
        self.assertEqual(bundle_list.version, 1)
        self.assertEqual(bundle_list.mode, "all")
        self.assertEqual(len(bundle_list.entries), 1)
        self.assertEqual(bundle_list.entries[0].id, "main")



# test_maintenance.py -- tests for porcelain maintenance
# Copyright (C) 2024 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for porcelain maintenance functions."""

import os
import shutil
import tempfile

from dulwich import porcelain
from dulwich.config import ConfigFile
from dulwich.objects import Blob, Commit, Tree
from dulwich.repo import Repo

from .. import TestCase


class PorcelainMaintenanceTestCase(TestCase):
    """Base class for porcelain maintenance tests."""

    def setUp(self):
        super().setUp()
        self.test_dir = tempfile.mkdtemp()
        self.addCleanup(self._cleanup_test_dir)
        self.repo = Repo.init(self.test_dir)
        self.addCleanup(self.repo.close)

    def _cleanup_test_dir(self):
        shutil.rmtree(self.test_dir)

    def _create_commit(self):
        """Helper to create a test commit."""
        blob = Blob.from_string(b"test content\n")
        self.repo.object_store.add_object(blob)

        tree = Tree()
        tree.add(b"testfile", 0o100644, blob.id)
        self.repo.object_store.add_object(tree)

        commit = Commit()
        commit.tree = tree.id
        commit.author = commit.committer = b"Test <test@example.com>"
        commit.author_time = commit.commit_time = 1000000000
        commit.author_timezone = commit.commit_timezone = 0
        commit.encoding = b"UTF-8"
        commit.message = b"Test commit"
        commit.parents = []
        self.repo.object_store.add_object(commit)
        self.repo.refs[b"refs/heads/master"] = commit.id
        return commit.id


class PorcelainMaintenanceTest(PorcelainMaintenanceTestCase):
    """Tests for porcelain.maintenance_run function."""

    def test_maintenance_run(self):
        """Test porcelain maintenance_run function."""
        self._create_commit()
        result = porcelain.maintenance_run(self.test_dir)
        self.assertIn("gc", result.tasks_succeeded)
        self.assertIn("commit-graph", result.tasks_succeeded)

    def test_maintenance_run_with_tasks(self):
        """Test porcelain maintenance_run with specific tasks."""
        result = porcelain.maintenance_run(self.test_dir, tasks=["pack-refs"])
        self.assertEqual(result.tasks_run, ["pack-refs"])
        self.assertEqual(result.tasks_succeeded, ["pack-refs"])


class MaintenanceRegisterTest(PorcelainMaintenanceTestCase):
    """Tests for maintenance register/unregister."""

    def setUp(self):
        super().setUp()
        # Set up a temporary HOME for testing global config
        self.temp_home = tempfile.mkdtemp()
        self.addCleanup(self._cleanup_temp_home)
        self.overrideHome(self.temp_home)

    def _cleanup_temp_home(self):
        shutil.rmtree(self.temp_home)

    def test_register_repository(self):
        """Test registering a repository for maintenance."""
        porcelain.maintenance_register(self.test_dir)

        # Verify repository was added to global config
        global_config_path = os.path.expanduser("~/.gitconfig")
        global_config = ConfigFile.from_path(global_config_path)

        repos = list(global_config.get_multivar((b"maintenance",), b"repo"))
        self.assertIn(self.test_dir.encode(), repos)

        # Verify strategy was set
        strategy = global_config.get((b"maintenance",), b"strategy")
        self.assertEqual(strategy, b"incremental")

        # Verify auto maintenance was disabled in repo
        repo_config = self.repo.get_config()
        auto = repo_config.get_boolean((b"maintenance",), b"auto")
        self.assertFalse(auto)

    def test_register_already_registered(self):
        """Test registering an already registered repository."""
        porcelain.maintenance_register(self.test_dir)
        # Should not error when registering again
        porcelain.maintenance_register(self.test_dir)

    def test_unregister_repository(self):
        """Test unregistering a repository."""
        # First register
        porcelain.maintenance_register(self.test_dir)

        # Then unregister
        porcelain.maintenance_unregister(self.test_dir)

        # Verify repository was removed from global config
        global_config_path = os.path.expanduser("~/.gitconfig")
        global_config = ConfigFile.from_path(global_config_path)

        try:
            repos = list(global_config.get_multivar((b"maintenance",), b"repo"))
            self.assertNotIn(self.test_dir.encode(), repos)
        except KeyError:
            # No repos registered, which is fine
            pass

    def test_unregister_not_registered(self):
        """Test unregistering a repository that is not registered."""
        with self.assertRaises(ValueError):
            porcelain.maintenance_unregister(self.test_dir)

    def test_unregister_not_registered_force(self):
        """Test unregistering with force flag."""
        # Should not error with force=True
        porcelain.maintenance_unregister(self.test_dir, force=True)



# test_cli_merge.py -- Tests for dulwich merge CLI command
# Copyright (C) 2024 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for dulwich merge CLI command."""

import importlib.util
import os
import stat
import tempfile
import unittest

from dulwich import porcelain
from dulwich.cli import main
from dulwich.objects import Blob, Commit, Tree
from dulwich.repo import Repo

from .. import DependencyMissing, TestCase


class CLIMergeTests(TestCase):
    """Tests for the dulwich merge CLI command."""

    def test_merge_fast_forward(self):
        """Test CLI merge with fast-forward."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Initialize repo
            porcelain.init(tmpdir)

            # Create initial commit
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Initial content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Initial commit")

            # Create a branch
            porcelain.branch_create(tmpdir, "feature")
            porcelain.checkout(tmpdir, "feature")

            # Add a file on feature branch
            with open(os.path.join(tmpdir, "file2.txt"), "w") as f:
                f.write("Feature content\n")
            porcelain.add(tmpdir, paths=["file2.txt"])
            porcelain.commit(tmpdir, message=b"Add feature")

            # Go back to master
            porcelain.checkout(tmpdir, "master")

            # Test merge via CLI
            old_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                with self.assertLogs("dulwich.cli", level="INFO") as cm:
                    ret = main(["merge", "feature"])
                    log_output = "\n".join(cm.output)

                self.assertEqual(ret, 0)  # Success
                self.assertIn("Merge successful", log_output)

                # Check that file2.txt exists
                self.assertTrue(os.path.exists(os.path.join(tmpdir, "file2.txt")))
            finally:
                os.chdir(old_cwd)

    def test_merge_with_conflicts(self):
        """Test CLI merge with conflicts."""

        # Check if merge3 module is available
        if importlib.util.find_spec("merge3") is None:
            raise DependencyMissing("merge3")

        with tempfile.TemporaryDirectory() as tmpdir:
            # Initialize repo
            porcelain.init(tmpdir)

            # Create initial commit
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Initial content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Initial commit")

            # Create a branch and modify file1
            porcelain.branch_create(tmpdir, "feature")
            porcelain.checkout(tmpdir, "feature")
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Feature content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Modify file1 in feature")

            # Go back to master and modify file1 differently
            porcelain.checkout(tmpdir, "master")
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Master content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Modify file1 in master")

            # Test merge via CLI - should exit with error
            old_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                with self.assertLogs("dulwich.cli", level="WARNING") as cm:
                    retcode = main(["merge", "feature"])
                    self.assertEqual(retcode, 1)
                    log_output = "\n".join(cm.output)

                self.assertIn("Merge conflicts", log_output)
                self.assertIn("file1.txt", log_output)
            finally:
                os.chdir(old_cwd)

    def test_merge_already_up_to_date(self):
        """Test CLI merge when already up to date."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Initialize repo
            porcelain.init(tmpdir)

            # Create initial commit
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Initial content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Initial commit")

            # Test merge via CLI
            old_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                with self.assertLogs("dulwich.cli", level="INFO") as cm:
                    ret = main(["merge", "HEAD"])
                    log_output = "\n".join(cm.output)

                self.assertEqual(ret, 0)  # Success
                self.assertIn("Already up to date", log_output)
            finally:
                os.chdir(old_cwd)

    def test_merge_no_commit(self):
        """Test CLI merge with --no-commit."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Initialize repo
            porcelain.init(tmpdir)

            # Create initial commit
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Initial content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Initial commit")

            # Create a branch
            porcelain.branch_create(tmpdir, "feature")
            porcelain.checkout(tmpdir, "feature")

            # Add a file on feature branch
            with open(os.path.join(tmpdir, "file2.txt"), "w") as f:
                f.write("Feature content\n")
            porcelain.add(tmpdir, paths=["file2.txt"])
            porcelain.commit(tmpdir, message=b"Add feature")

            # Go back to master and add another file
            porcelain.checkout(tmpdir, "master")
            with open(os.path.join(tmpdir, "file3.txt"), "w") as f:
                f.write("Master content\n")
            porcelain.add(tmpdir, paths=["file3.txt"])
            porcelain.commit(tmpdir, message=b"Add file3")

            # Test merge via CLI with --no-commit
            old_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                with self.assertLogs("dulwich.cli", level="INFO") as cm:
                    ret = main(["merge", "--no-commit", "feature"])
                    log_output = "\n".join(cm.output)

                self.assertEqual(ret, 0)  # Success
                self.assertIn("not committing", log_output)

                # Check that files are merged
                self.assertTrue(os.path.exists(os.path.join(tmpdir, "file2.txt")))
                self.assertTrue(os.path.exists(os.path.join(tmpdir, "file3.txt")))
            finally:
                os.chdir(old_cwd)

    def test_merge_no_ff(self):
        """Test CLI merge with --no-ff."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Initialize repo
            porcelain.init(tmpdir)

            # Create initial commit
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Initial content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Initial commit")

            # Create a branch
            porcelain.branch_create(tmpdir, "feature")
            porcelain.checkout(tmpdir, "feature")

            # Add a file on feature branch
            with open(os.path.join(tmpdir, "file2.txt"), "w") as f:
                f.write("Feature content\n")
            porcelain.add(tmpdir, paths=["file2.txt"])
            porcelain.commit(tmpdir, message=b"Add feature")

            # Go back to master
            porcelain.checkout(tmpdir, "master")

            # Test merge via CLI with --no-ff
            old_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                with self.assertLogs("dulwich.cli", level="INFO") as cm:
                    ret = main(["merge", "--no-ff", "feature"])
                    log_output = "\n".join(cm.output)

                self.assertEqual(ret, 0)  # Success
                self.assertIn("Merge successful", log_output)
                self.assertIn("Created merge commit", log_output)
            finally:
                os.chdir(old_cwd)

    def test_merge_with_message(self):
        """Test CLI merge with custom message."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Initialize repo
            porcelain.init(tmpdir)

            # Create initial commit
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Initial content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Initial commit")

            # Create a branch
            porcelain.branch_create(tmpdir, "feature")
            porcelain.checkout(tmpdir, "feature")

            # Add a file on feature branch
            with open(os.path.join(tmpdir, "file2.txt"), "w") as f:
                f.write("Feature content\n")
            porcelain.add(tmpdir, paths=["file2.txt"])
            porcelain.commit(tmpdir, message=b"Add feature")

            # Go back to master and add another file
            porcelain.checkout(tmpdir, "master")
            with open(os.path.join(tmpdir, "file3.txt"), "w") as f:
                f.write("Master content\n")
            porcelain.add(tmpdir, paths=["file3.txt"])
            porcelain.commit(tmpdir, message=b"Add file3")

            # Test merge via CLI with custom message
            old_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                with self.assertLogs("dulwich.cli", level="INFO") as cm:
                    ret = main(["merge", "-m", "Custom merge message", "feature"])
                    log_output = "\n".join(cm.output)

                self.assertEqual(ret, 0)  # Success
                self.assertIn("Merge successful", log_output)
            finally:
                os.chdir(old_cwd)

    def test_octopus_merge_three_branches(self):
        """Test CLI octopus merge with three branches."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Initialize repo
            porcelain.init(tmpdir)

            # Create initial commit with three files
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("File 1 content\n")
            with open(os.path.join(tmpdir, "file2.txt"), "w") as f:
                f.write("File 2 content\n")
            with open(os.path.join(tmpdir, "file3.txt"), "w") as f:
                f.write("File 3 content\n")
            porcelain.add(tmpdir, paths=["file1.txt", "file2.txt", "file3.txt"])
            porcelain.commit(tmpdir, message=b"Initial commit")

            # Create branch1 and modify file1
            porcelain.branch_create(tmpdir, "branch1")
            porcelain.checkout(tmpdir, "branch1")
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Branch1 modified file1\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Branch1 modifies file1")

            # Create branch2 and modify file2
            porcelain.checkout(tmpdir, "master")
            porcelain.branch_create(tmpdir, "branch2")
            porcelain.checkout(tmpdir, "branch2")
            with open(os.path.join(tmpdir, "file2.txt"), "w") as f:
                f.write("Branch2 modified file2\n")
            porcelain.add(tmpdir, paths=["file2.txt"])
            porcelain.commit(tmpdir, message=b"Branch2 modifies file2")

            # Create branch3 and modify file3
            porcelain.checkout(tmpdir, "master")
            porcelain.branch_create(tmpdir, "branch3")
            porcelain.checkout(tmpdir, "branch3")
            with open(os.path.join(tmpdir, "file3.txt"), "w") as f:
                f.write("Branch3 modified file3\n")
            porcelain.add(tmpdir, paths=["file3.txt"])
            porcelain.commit(tmpdir, message=b"Branch3 modifies file3")

            # Go back to master and octopus merge all three branches via CLI
            porcelain.checkout(tmpdir, "master")
            old_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                with self.assertLogs("dulwich.cli", level="INFO") as cm:
                    ret = main(["merge", "branch1", "branch2", "branch3"])
                    log_output = "\n".join(cm.output)

                self.assertEqual(ret, 0)  # Success
                self.assertIn("Octopus merge successful", log_output)

                # Check that all modifications are present
                with open(os.path.join(tmpdir, "file1.txt")) as f:
                    self.assertEqual(f.read(), "Branch1 modified file1\n")
                with open(os.path.join(tmpdir, "file2.txt")) as f:
                    self.assertEqual(f.read(), "Branch2 modified file2\n")
                with open(os.path.join(tmpdir, "file3.txt")) as f:
                    self.assertEqual(f.read(), "Branch3 modified file3\n")
            finally:
                os.chdir(old_cwd)

    def test_octopus_merge_with_conflicts(self):
        """Test CLI octopus merge with conflicts."""

        # Check if merge3 module is available
        if importlib.util.find_spec("merge3") is None:
            raise DependencyMissing("merge3")

        with tempfile.TemporaryDirectory() as tmpdir:
            # Initialize repo
            porcelain.init(tmpdir)

            # Create initial commit
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Initial content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Initial commit")

            # Create branch1 and modify file1
            porcelain.branch_create(tmpdir, "branch1")
            porcelain.checkout(tmpdir, "branch1")
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Branch1 content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Branch1 modifies file1")

            # Create branch2 and modify file1 differently
            porcelain.checkout(tmpdir, "master")
            porcelain.branch_create(tmpdir, "branch2")
            porcelain.checkout(tmpdir, "branch2")
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Branch2 content\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Branch2 modifies file1")

            # Go back to master and try octopus merge via CLI - should fail
            porcelain.checkout(tmpdir, "master")
            old_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                with self.assertLogs("dulwich.cli", level="WARNING") as cm:
                    ret = main(["merge", "branch1", "branch2"])
                    log_output = "\n".join(cm.output)

                self.assertEqual(ret, 1)  # Error
                self.assertIn("Octopus merge failed", log_output)
                self.assertIn("refusing to merge with conflicts", log_output)
            finally:
                os.chdir(old_cwd)

    def test_octopus_merge_no_commit(self):
        """Test CLI octopus merge with --no-commit."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Initialize repo
            porcelain.init(tmpdir)

            # Create initial commit
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("File 1 content\n")
            with open(os.path.join(tmpdir, "file2.txt"), "w") as f:
                f.write("File 2 content\n")
            porcelain.add(tmpdir, paths=["file1.txt", "file2.txt"])
            porcelain.commit(tmpdir, message=b"Initial commit")

            # Create branch1 and modify file1
            porcelain.branch_create(tmpdir, "branch1")
            porcelain.checkout(tmpdir, "branch1")
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("Branch1 modified file1\n")
            porcelain.add(tmpdir, paths=["file1.txt"])
            porcelain.commit(tmpdir, message=b"Branch1 modifies file1")

            # Create branch2 and modify file2
            porcelain.checkout(tmpdir, "master")
            porcelain.branch_create(tmpdir, "branch2")
            porcelain.checkout(tmpdir, "branch2")
            with open(os.path.join(tmpdir, "file2.txt"), "w") as f:
                f.write("Branch2 modified file2\n")
            porcelain.add(tmpdir, paths=["file2.txt"])
            porcelain.commit(tmpdir, message=b"Branch2 modifies file2")

            # Go back to master and octopus merge with --no-commit
            porcelain.checkout(tmpdir, "master")
            old_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                with self.assertLogs("dulwich.cli", level="INFO") as cm:
                    ret = main(["merge", "--no-commit", "branch1", "branch2"])
                    log_output = "\n".join(cm.output)

                self.assertEqual(ret, 0)  # Success
                self.assertIn("not committing", log_output)

                # Check that files are merged
                with open(os.path.join(tmpdir, "file1.txt")) as f:
                    self.assertEqual(f.read(), "Branch1 modified file1\n")
                with open(os.path.join(tmpdir, "file2.txt")) as f:
                    self.assertEqual(f.read(), "Branch2 modified file2\n")
            finally:
                os.chdir(old_cwd)

    def test_merge_invalid_path_returns_error(self):
        """An invalid path in the merged tree returns exit code 1, not a crash."""
        with tempfile.TemporaryDirectory() as tmpdir:
            with Repo.init(tmpdir) as repo:
                repo.get_config().set((b"core",), b"protectNTFS", b"true")
                repo.get_config().write_to_path()

                blob = Blob.from_string(b"base\n")
                repo.object_store.add_object(blob)
                base_tree = Tree()
                base_tree.add(b"ok.txt", stat.S_IFREG | 0o644, blob.id)
                repo.object_store.add_object(base_tree)
                base = Commit()
                base.tree = base_tree.id
                base.author = base.committer = b"Test <test@example.com>"
                base.author_time = base.commit_time = 0
                base.author_timezone = base.commit_timezone = 0
                base.message = b"base\n"
                repo.object_store.add_object(base)
                repo.refs[b"refs/heads/master"] = base.id
                repo.refs.set_symbolic_ref(b"HEAD", b"refs/heads/master")
                repo.get_worktree().reset_index(base_tree.id)

                payload = Blob.from_string(b"payload\n")
                repo.object_store.add_object(payload)
                attack_tree = Tree()
                attack_tree.add(b"ok.txt", stat.S_IFREG | 0o644, blob.id)
                # ``git~1`` is the NTFS 8.3 short-name alias for ``.git``.
                attack_tree.add(b"git~1", stat.S_IFREG | 0o755, payload.id)
                repo.object_store.add_object(attack_tree)
                attack = Commit()
                attack.tree = attack_tree.id
                attack.parents = [base.id]
                attack.author = attack.committer = b"Test <test@example.com>"
                attack.author_time = attack.commit_time = 0
                attack.author_timezone = attack.commit_timezone = 0
                attack.message = b"attack\n"
                repo.object_store.add_object(attack)

            old_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                with self.assertLogs("dulwich.cli", level="ERROR"):
                    ret = main(["merge", attack.id.decode()])
                self.assertEqual(ret, 1)
            finally:
                os.chdir(old_cwd)


if __name__ == "__main__":
    unittest.main()



# test_pack.py -- property tests for pack.py
# Copyright (C) 2026 The Dulwich contributors
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Property tests for pack helpers."""

import os
from collections.abc import Iterable

from hypothesis import example, given, settings
from hypothesis import strategies as st

from dulwich.errors import ApplyDeltaError
from dulwich.pack import _create_delta_py, _delta_encode_size, apply_delta, create_delta
from tests import TestCase


def _delta_to_bytes(delta: bytes | Iterable[bytes]) -> bytes:
    if isinstance(delta, bytes):
        return delta
    return b"".join(delta)


settings.register_profile(
    "deterministic", max_examples=50, deadline=None, derandomize=True
)
settings.register_profile("ci", max_examples=50, deadline=None, derandomize=True)
settings.register_profile(
    "local-deep", max_examples=1000, deadline=None, derandomize=True
)
settings.load_profile(os.environ.get("HYPOTHESIS_PROFILE", "deterministic"))

byte_strings = st.binary(max_size=256)


@st.composite
def delta_pairs(draw) -> tuple[bytes, bytes]:
    """Generate byte pairs with some shared content for delta copies."""
    prefix = draw(st.binary(max_size=128))
    suffix = draw(st.binary(max_size=128))
    base_middle = draw(st.binary(max_size=128))
    target_middle = draw(st.binary(max_size=128))
    return prefix + base_middle + suffix, prefix + target_middle + suffix


@st.composite
def bounded_delta_inputs(draw) -> tuple[bytes, bytes]:
    """Generate arbitrary delta op streams with bounded output sizes."""
    base = draw(byte_strings)
    dest_size = draw(st.integers(min_value=0, max_value=512))
    ops = draw(st.binary(max_size=256))
    delta = _delta_encode_size(len(base)) + _delta_encode_size(dest_size) + ops
    return base, delta


byte_pairs = st.one_of(st.tuples(byte_strings, byte_strings), delta_pairs())


class PackPropertyTests(TestCase):
    """Property tests for pack helpers."""

    @given(byte_pairs)
    @example((b"", b""))
    @example((b"", b"Z" * 8192))
    @example((b"Z" * 8192, b"Z" * 8192))
    @example((b"Z" * 70000 + b"a", b"Z" * 70000 + b"b"))
    def test_create_delta_roundtrip(self, pair: tuple[bytes, bytes]) -> None:
        """Check that generated deltas apply back to the target."""
        base, target = pair
        delta = _delta_to_bytes(create_delta(base, target))
        self.assertEqual(target, b"".join(apply_delta(base, delta)))

    @given(byte_pairs)
    @example((b"", b""))
    @example((b"", b"Z" * 8192))
    @example((b"Z" * 8192, b"Z" * 8192))
    @example((b"Z" * 70000 + b"a", b"Z" * 70000 + b"b"))
    def test_create_delta_py_roundtrip(self, pair: tuple[bytes, bytes]) -> None:
        """Check that pure Python generated deltas apply to the target."""
        base, target = pair
        delta = _delta_to_bytes(_create_delta_py(base, target))
        self.assertEqual(target, b"".join(apply_delta(base, delta)))

    @given(bounded_delta_inputs())
    @example((b"", b"\x00\x01\x01"))
    def test_apply_delta_only_raises_apply_delta_error(
        self, base_and_delta: tuple[bytes, bytes]
    ) -> None:
        """Check that malformed deltas use the delta error type."""
        base, delta = base_and_delta
        try:
            apply_delta(base, delta)
        except ApplyDeltaError:
            pass



# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later

import sys
from io import BytesIO

import atheris

with atheris.instrument_imports():
    from dulwich.bitmap import EWAHBitmap, read_bitmap_file


def TestOneInput(data) -> int | None:
    # Exercise the standalone EWAH decoder as well as the full bitmap file
    # reader, both of which parse untrusted input.
    try:
        EWAHBitmap(data)
    except ValueError:
        pass

    try:
        read_bitmap_file(BytesIO(data))
    except ValueError:
        pass


def main() -> None:
    atheris.Setup(sys.argv, TestOneInput)
    atheris.Fuzz()


if __name__ == "__main__":
    main()



# test_lfs_server.py -- Tests for the LFS server
# Copyright (C) 2024 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for the LFS server."""

import http.client
import os
import shutil
import tempfile
import threading

from dulwich.lfs_server import run_lfs_server

from . import TestCase


class LFSServerOidValidationTests(TestCase):
    """Tests that the server rejects object ids that escape the store."""

    def setUp(self) -> None:
        super().setUp()
        self.base = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.base)
        lfs_dir = os.path.join(self.base, "lfs")
        self.server, self.url = run_lfs_server(lfs_dir=lfs_dir)
        thread = threading.Thread(target=self.server.serve_forever)
        thread.daemon = True
        thread.start()

        def stop() -> None:
            self.server.shutdown()
            thread.join()
            self.server.server_close()

        self.addCleanup(stop)
        host, port = self.server.server_address[:2]
        self.host = host
        self.port = port

    def _get(self, path: str) -> tuple[int, bytes]:
        # Use a raw connection so the request path is sent verbatim and not
        # normalized/percent-encoded by a higher level client.
        conn = http.client.HTTPConnection(self.host, self.port)
        try:
            conn.request("GET", path)
            resp = conn.getresponse()
            return resp.status, resp.read()
        finally:
            conn.close()

    def test_valid_object_is_served(self) -> None:
        oid = self.server.lfs_store.write_object([b"hello"])
        status, body = self._get(f"/objects/{oid}")
        self.assertEqual(200, status)
        self.assertEqual(b"hello", body)

    def test_traversing_oid_is_rejected(self) -> None:
        # ``_sha_path`` joins objects/<oid[:2]>/<oid[2:4]>/<oid>, so an oid of
        # ".." for the first two components climbs out of the store. Plant a
        # file where ``....`` would resolve and confirm it is not served.
        secret = os.path.join(self.base, "....")
        with open(secret, "wb") as f:
            f.write(b"secret")
        status, body = self._get("/objects/....")
        self.assertEqual(404, status)
        self.assertNotIn(b"secret", body)

    def test_backslash_oid_is_rejected(self) -> None:
        # On Windows the backslash is a path separator, so this would traverse;
        # it must be refused on every platform.
        status, _ = self._get("/objects/..\\..\\..\\secret")
        self.assertEqual(404, status)

    def test_non_hex_oid_is_rejected(self) -> None:
        status, _ = self._get("/objects/" + "z" * 64)
        self.assertEqual(404, status)
